class Entity{

  // il nome del nodo
  String name;
  
  // il tipo del nodo (people, org, missing scene, knowledege sharing....)
  String type;
  
  // i "figli" del nodo (quindi quelli a cui e' legato con una relazione padre-figlio)
  String[] children;
  
  // indica se il nodo e' correntemente visibile (inizialmente impostato
  // tramite la configurazione del tipo di nodo)
  boolean isVisible;
  
  // tre vettori della posizione, velocita' e accelerazione del nodo
  PVector position, speed, acceleration;
  
  // configurazione visuale del nodo
  NodeConfiguration configuration;
  
  
  // e' il costruttore che viene usato caricando la base di dati
  public Entity(String n, String t, String[] c){
  
    name = n;
    type = t;
    children = c;
    
    // imposta una configurazione di default
    configuration = new NodeConfiguration();
    
    // usando la configurazione di default imposta la visibilita' iniziale del nodo
    isVisible = configuration.visualizeInitially;
    
    // inizializza i 3 vettori di posizione/velocita'/accelerazione a 0
    position = new PVector();
    speed = new PVector(0,0);
    acceleration = new PVector(0,0);
    
    // genera una posizione random sullo schermo
    position.x = random( 0,width );
    position.y = random( 0,height);
    
    // per evitare errori --> se il nodo non ha figli, sostituisce children=null con
    // con un "children = new String[0]", ovvero un insieme vuoto 
    if(children==null){
      children = new String[0];
    }
    
  }
  
  
  
  // preserva e ripristina il contesto grafico
  // crea delle forze di frizione, per stabilizzare la visualizzazione
  // applica la accelerazione alla velocita' e la velocita' alla posizione
  // usa la configurazione del nodo per disegnarlo
  void draw(){
  
    // preserva il contesto grafico
    pushMatrix();
    
    // frizione da velocita'
    // mi creo un vettore
    PVector frictionSpeed = new PVector();
    // ci copio dentro il vettore velocita'
    frictionSpeed.x = speed.x;
    frictionSpeed.y = speed.y;
    // ne inverto la direzione
    frictionSpeed.mult( -1 );
    // lo moltiplico per una costante di frizione
    // (potrebbe essere interessante da rendere configurabile: se la aumento
    // aumenta la frizione di questo tipo)
    frictionSpeed.mult( 0.08 );
    // aggiungo la forza di frizione di questo tipo
    addForce( frictionSpeed );
    
    //frizione da accelerazione 
    // calcola nello stesso modo, ma partendo dal vettore accelerazione
    PVector frictionAcceleration = new PVector();
    frictionAcceleration.x = acceleration.x;
    frictionAcceleration.y = acceleration.y;
    frictionAcceleration.mult( -0.08 );
    addForce( frictionAcceleration );
    
    // aggiorno la velocita' attraverso tutte le accelerazioni (forze) accumulate
    // fino a questo punto
    speed.add( acceleration );
    
    // aggiorno la posizione aggiungendo la velocita' per far spostare il nodo
    position.add( speed );
    
    /*
    // qui controlliamo se le coordinate dell'entity sono fuori dallo schermo
    // se e' cosi' le riportiamo dentro il bordo
    if(position.x<0){ position.x = 0; }
    if(position.x>width){ position.x = width; }
    if(position.y<0){ position.y = 0; }
    if(position.y>height){ position.y = height; }
    */
    
    
    // l'accelerazione, come le forze, e' un impulso
    // --> l'usata e quindi posso metterla uguale a 0
    // (visto che e' un impulso (si usa solo una volta) 
    acceleration.x = 0;
    acceleration.y = 0;
    
    //move to the position of the node
    translate(position.x , position.y);
    
    // usiamo la configurazione visuale per stabilire il colore
    fill( configuration.r , configuration.g , configuration.b ); 
    noStroke();
    
    // se e' stata configurata una icona
    if(configuration.icon!=null){
      // disegno l'icona e la centro usando la sua larghezza/altezza
      image(configuration.icon,  -configuration.icon.width/2 +globalOffsetX , -configuration.icon.height/2 + globalOffsetY);
    } else {
      // altrimenti disegno un cerchio
      ellipse(0 + globalOffsetX,0 + globalOffsetY ,5,5);
    }
    
    // disegno l'etichetta
    // nota: font, dimensione e colore dell'etichetta potrebbero essere resi configurabili
    // nel file di configurazione dei nodi
    // imposto il font
    textFont(font1 , 10 );
    // imposto il colore del font
    fill(30,30,30);
    noStroke();
    
    float larghezzaEtichetta = textWidth( name );
    
    if( (position.x + larghezzaEtichetta +10) > width){
       // dobbiamo disenare l'etichetta a sinistra
       // disegno il nome del nodo, a 10px di distanza versto sx
      text(name, -10 -larghezzaEtichetta +globalOffsetX,0 + globalOffsetY );
    } else {
      // disegno il nome del nodo, a 10px di distanza versto dx
      text(name, 10 + globalOffsetX ,0 + globalOffsetY );
    }
    
    
    
    // ripristina il contesto grafico
    popMatrix();
    
    
  }

  
  // e' il metodo con cui si aggiunge una forza al nodo
  void addForce( PVector v ){
  
    // la forza si traduce in accelerazione e quindi va a modificare direttamente
    // il vettore dell'accelerazione
    acceleration.add( v );
    
  }


}
