import g4p_controls.*;

// in this section we define the global variables, the variables
// which are accessible from everywhere in the code

// array lists containing all the collections needed
// in the program (nodes, relations, configurations...)
// each ArrayList is defined for a certain class so that
// if you want to day that the nodes array contains element
// of the Entity type, you write:
// ArrayList<Entity> nodes;
// this means that each cell of the arraylist is of type Entity.
ArrayList<Entity> nodes;
ArrayList<Relation> relations;
ArrayList<NodeConfiguration> nodeConfigurations;
ArrayList<RelationConfiguration> relationConfigurations;


float globalOffsetX, globalOffsetY;

// this is a font we will use for labels
PFont font1;

String filteredRelations;

// lo slider e la variabile che controllano la soglia di distanza tra padri e figli 
GCustomSlider sdr1;
float sogliaDistanzaPadreFiglio = 150;

// lo slider e la variabile che controllano la soglia di distanza generale 
GCustomSlider sdr2;
float sogliaDistanzaRelazioni = 150;


ArrayList<String> listaTagRelazioni;


/*

how to make sliders:

  //=============================================================
  // Simple default slider,
  // constructor is `Parent applet', the x, y position and length
  sdr1 = new GCustomSlider(this, 20, 20, 260, 50, null);
  // show          opaque  ticks value limits
  sdr1.setShowDecor(false, true, true, true);
  sdr1.setNbrTicks(5);
  sdr1.setLimits(40, -100, 100);
  
  // and add a function like this to handle slider events (value changed):
  public void handleSliderEvents(GValueControl slider, GEvent event){
    println("integer value:" + slider.getValueI() + " float value:" + slider.getValueF());
  }
  
  
  // how to make text boxes
  txf1 = new GTextField(this, 10, 10, 200, 20);
  txf1.tag = "txf1";
  txf1.setDefaultText("Text field 1"); 
  
  
  // and add a function like this to handle text events
  public void handleTextEvents(GEditableTextControl tc, GEvent event) { 
  System.out.print("\n" + tc.tag + "   Event type: ");
  switch(event) {
  case CHANGED:
    System.out.println("CHANGED");
    break;
  case SELECTION_CHANGED:
    System.out.println("SELECTION_CHANGED");
    System.out.println(tc.getSelectedText() + "\n");
    break;
  case ENTERED:
    System.out.println("ENTER KEY TYPED");
    System.out.println(tc.getSelectedText() + "\n");
    break;
  default:
    System.out.println("UNKNOWN");
  }
}
  
  

*/


void setup(){

  size(1000,500);
  
  font1 = loadFont( "font1.vlw" );
  
  globalOffsetX = 0;
  globalOffsetY = 0;
  
  listaTagRelazioni = new ArrayList<String>();
  
  
  // inizializziamo lo slider 1
  // new GCustomSlider(this, x angolo alto-sx, y angolo alto-sx, x angolo basso-dx, y angolo basso-dx, null);
  sdr1 = new GCustomSlider(this, 20, 20, 100, 40, null);
  sdr1.tag = "sogliaDistanzaPadreFiglio";
  // show          opaque  ticks value limits
  sdr1.setShowDecor(false, false, true, true);
  sdr1.setNbrTicks(5);
  // imposta i limiti (escursione) dello slider: valore iniziale, minimo e massimo 
  sdr1.setLimits(sogliaDistanzaPadreFiglio, 30, 220);
  
  
  
  // inizializziamo lo slider 2
  // new GCustomSlider(this, x angolo alto-sx, y angolo alto-sx, x angolo basso-dx, y angolo basso-dx, null);
  sdr2 = new GCustomSlider(this, 20, 50, 100, 70, null);
  sdr2.tag = "sogliaDistanzaRelazioni";
  // show          opaque  ticks value limits
  sdr2.setShowDecor(false, false, true, true);
  sdr2.setNbrTicks(5);
  // imposta i limiti (escursione) dello slider: valore iniziale, minimo e massimo 
  sdr2.setLimits(sogliaDistanzaRelazioni, 80, 500);
  
  
  
  filteredRelations = "communication";
  
  // loading of the relations configurations
  
  // con loadStrings leggiamo un file di testo, che ci viene
  // restituito sotto forma di array di Stringhe (una per riga)
  String[] dati4 = loadStrings("relationsConfiguration.csv" );
  
  //inizializziamo l'arraylist per contenere le configurazioni
  relationConfigurations = new ArrayList<RelationConfiguration>();
  
  
  // come funziona un ciclo for:
  // for( inizializzazione ; condizione ; incremento ){ qualcosa da eseguire }
  // inizializzazione: viene eseguita una sola volta all'inizio (da dove inizio a contare?)
  // condizione: dice fino a quando contare; finche' la condizione e' vera, continuo a contare
  // incremento: come conto (incremento di 1, per esempio, in questo caso)
  // in questo caso:
  // conto da 0 al numero delle righe nel file di configurazione (dati4.length)
  for( int i=0; i<dati4.length ; i=i+1)
  {
    // prendo la i-esima riga del file di configurazione
    String line = dati4[i];
    
    // se il primo carattere [charAt(0)] della riga e' un # vuol dire
    // che e' un commento quindi lo salto
    if(line.charAt( 0 )!='#'){
    
      // non e' un commento, quindi lo elaboro
      
      // println("line " + i + " is:" + line);
    
      // il comando split prende una stringa (in questo caso la riga del
      // file di configurazione) e la spezza secondo un certo delimitatore
      // (in questo caso il carattere | )
      String[] parts = split( line , '|' );
    
      // parts[0] --> name of the type
      // parts[1] --> r,g,b colors
      // parts[2] --> name of a image file to be found in the "data" folder
      // parts[3] --> a number, which represents the offset in %
      
      // elaboriamo "r,g,b"
      // li splittiamo con la virgola
      String[] numbers = split(parts[1],",");
      // per evitare errori impostiamo anche un valore di default (in questo caso
      // corrisponde al bianco)
      int r = 255;
      int g = 255;
      int b = 255;
      // se ci sono esattamente tre numeri ottenuti splittando r,g,b,li prendo
      if(numbers!=null && numbers.length==3){
        //r e' uguale al primo, g al secondo e b al terzo
        // la funzione "int( qualcosa )" prende quel "qualcosa" e cerca di trasformarlo
        // in un numero intero
        // es:
        // qualcosa=12 --> il numero 12
        // qualcosa=24.6 --> il numero 24
        // qualcosa=mela --> il numero 0
        r = int( numbers[0] );
        g = int( numbers[1] );
        b = int( numbers[2] );
      }
      
      // facciamo la stessa cosa con l'offset che abbiamo nella 4a posizione
      // di ogni riga
      // usiamo "float( qualcosa )" che nello stesso modo di "int(qualcosa)" solo che
      // restituisce un numero con virgola
      // es:
      // qualcosa=0.5 --> il numero 0.5
      // qualcosa=-234 --> il numero -234.0
      // qualcosa=mela --> il numero 0.0
      float o = float( parts[3] )/100;
      
      // ho tutti i pezzi che mi servono per inizializzare un nuovo elemento
      // di tipo "RelationConfiguration" --> lo inizializzo 
      RelationConfiguration rc = new RelationConfiguration( parts[0], r, g, b, parts[2], o );
      
      // dopo averlo inizializzato, lo aggiungo agli altri
      relationConfigurations.add(rc);
    }
  }
  
  // loading of the relations configurations - END
  
  
  
  
  
  
  
  
  // loading of the nodes configurations
  
  String[] dati3 = loadStrings("nodesConfiguration.csv" );
  
  nodeConfigurations = new ArrayList<NodeConfiguration>();
  
  for(
        int i=0; i<dati3.length ; i=i+1
      )
  {
    String line = dati3[i];
    
    if(line.charAt( 0 )!='#'){
    
      //println("line " + i + " is:" + line);
    
      String[] parts = split( line , '|' );
    
      // part[0] --> name of the type
      // part[1] --> r,g,b colors
      // part[2] --> name of a image file to be found in the "data" folder
      // part[3] --> y or n to be transformed into true or false
      
      String[] numbers = split(parts[1],",");
      int r = 255;
      int g = 255;
      int b = 255;
      if(numbers!=null && numbers.length==3){
        r = int( numbers[0] );
        g = int( numbers[1] );
        b = int( numbers[2] );
      }
      
      // in parts[3] ho un valore "y" o "n" (yes/no)
      // che devo trasformare in un valore vero/falso (un boolean)
      // --> definisco un valore di default (quindi presuppongo
      // che se non specifico nulla, il tipo di nodo sara' visibile
      boolean isInitially = true;
      
      // controllo se parts[3] inizia con "y"
      if(parts[3].startsWith( "y" )){
        // vuol dire che il tipo di nodo sara' visualizzabile sin dall'inizio
        isInitially = true;
      } else {
        // vuol dire che non e' visibile sin dall'inizio
        isInitially = false;
      }
      
      NodeConfiguration nc = new NodeConfiguration( parts[0], r, g, b, parts[2], isInitially );
      nodeConfigurations.add(nc);
      
    }
    
  }
  
  // loading of the nodes configurations - END
  
  
  
  
  
  // loading of the relations
  
  String[] dati2 = loadStrings("relations.csv" );
  relations = new ArrayList<Relation>();
  for(
        int i=0; i<dati2.length ; i=i+1
      )
  {
    String line = dati2[i];
    
    println("line " + i + " is:" + line);
    
    String[] parts = split( line , '|' );
    
    String[] tags = split( parts[3], ',');
    
    
    for(int jj = 0; jj<tags.length; jj++){
      listaTagRelazioni.add( tags[jj] );
    }
    
    Relation r = new Relation( parts[0] , parts[1], parts[2], tags , parts[4]  );
    
    
    // in parts[2] we have the type of the relation
    // vogliamo capire se tra le configurazioni ce n'e' una 
    // che si applica al tipo di relazione corrente
    // quindi faccio un ciclo su relationConfigurations 
    for(int j=0; j<relationConfigurations.size(); j++){
      
      // prendo una per volta le configurazioni impostate
      RelationConfiguration rc = relationConfigurations.get( j );
      
      // vedo se il tipo e' uguale a quello della relazione attuale (parts[2])
      if(rc.type.equals( parts[2] )){
        // we have found the right configuration
        // quindi la imposto per la relazione attuale
        r.configuration = rc;
      }
    }
    
    
    
    relations.add(r);
    
  }
  
  
  // loading of the relations - END
  
  
  
  
  
  // loading of the nodes

  String[] dati = loadStrings("datiOrgModel.csv" );
  
  nodes = new ArrayList<Entity>();
  
  for(
        int i=0; 
        i<dati.length ; 
        i=i+1
      )
  {
    String line = dati[i];
    
    //println("line " + i + " is:" + line);
    
    String[] parts = split( line , '|' );
    
    //
    
    String[] children = split( parts[2], ',');
    
    Entity newNode = new Entity(  parts[0], parts[1] , children  );
    
    // in parts[1] we have the type of the node
    // discorso praticamente uguale a quello della configurazione
    // delle relazioni, nel blocco precedente
    for(int j=0; j<nodeConfigurations.size(); j++){
      NodeConfiguration nc = nodeConfigurations.get( j );
      if(nc.type.equals( parts[1] )){
        // we have found the right configuration
        newNode.configuration = nc;
        newNode.isVisible = nc.visualizeInitially;
      }
    }
    
    nodes.add( newNode );
    
  }

  // loading of the nodes - END
  
  /*
  for(int i=0; i<listaTagRelazioni.size(); i++){
    println(listaTagRelazioni.get( i ));
  }
  */


} // end of setup()





void draw(){
  
  // imposto lo sfondo
  background(220 , 220 ,220  );
  
  
  // inizializzo la drawing cache
  // ci metto man mano le immagini/icone che voglio che appiano sopra tutte le altre
  // e che poi disegnero' tutte insieme alla fine del draw() 
  ArrayList<DrawingCache> cache = new ArrayList<DrawingCache>();
  
  // chiamo la funzione che gestisce il mouse (definita piu' in basso)
  handleMouse();
  
  
  // crea la forza di repulsione se i nodi sono troppo vicini
  // prendo tutte le coppie di nodi e le confronto:
  // faccio due cicli sui nodi per vedere se sono troppo vicini
  // questo e' il primo ciclo
  for(int i = 0; i<nodes.size(); i++){
    // prendo il primo nodo come l'i-esimo dell'insieme dei nodi
    Entity e1 = nodes.get( i );
    
    // lo elaboro solo se e' visibile
    if(e1.isVisible){
      
    // questo e' il secondo ciclo
    // in questo ciclo inizio a contare da (i+1) perche' i precenti
    // li ho gia' confrontati 
    for(int j = i+1; j<nodes.size(); j++){
      
      // prendo il secondo nodo come il j-esimo dell'insieme dei nodi
      Entity e2 = nodes.get( j );  
      
      // lo elaboro solo se e' visibile
      if(e2.isVisible){
        
        
      // elaborazione
      // calcolo il vettore differenza tra le due posizioni dei nodi scelti
      
      // inizializzo un vettore vuoto
      PVector distanza = new PVector();
      
      // ci copio dentro la posizione del primo nodo
      distanza.x = e1.position.x;
      distanza.y = e1.position.y;
      
      // gli sottraggo la posizione del secondo nodo
      distanza.sub( e2.position );
      
      // --> ottengo in "distanza" il vettore differenza
      // che collega i due nodi
      
      // ne misuro la lunghezza con il comando "mag()"
      float lunghezza = distanza.mag();
      
      // se la lunghezza e' minore di 50 --> che devo farli allontanare
      // perche' sono troppo vicini
      // (questo e' uno dei parametri che potremmo includere in una interfaccia
      // grafica: uno slider che fa stare i nodi piu' vicini o lontani )
      // max(A,B) --> restituisce il valore massimo tra A e B;
      // in questo caso lo uso per non far andare sotto 0 "sogliaDistanzaRelazioni-100" 
      if(lunghezza<  max(0,sogliaDistanzaRelazioni-100)  ){
        
        // vuol dire che sono troppo vicini
        // --> applico una forza repellente
        
        // se lascio il vettore cosi' com'e' --> i due nodi sarebbero respinti
        // istantaneamente --> lo divido per una costante (0.01) per rallentare
        // la forza repulsiva (anche questo e' un parametro che potrebbe essere
        // reso configurabile: se aumento 0.01 aumenta la repulsione)
        distanza.mult( 0.01 );
        
        // applico una forza a e1
        e1.addForce( distanza );
        
        // inverto la forza moltiplicandola per -1 (inverto la direzione del vettore)
        distanza.mult( -1 );
        
        // applico la forza invertita a e2
        e2.addForce( distanza );
        
      }
      
      }//finisce if(e2.isVisible){
    }
    }// finisce if(e1.isVisible){
  }
  // crea la forza di repulsione se i nodi sono troppo vicini - END
  
  
  
  
  
  // crea le forze di attrazione tra i nodi ed i loro children
  // fare un loop attraverso i nodi
  for(int i = 0; i<nodes.size(); i++){
    
    // prendo il nodo i-esimo
    Entity e1 = nodes.get( i );
    
    // se il nodo e' visibile
    if(e1.isVisible){
    
    // prendi i suoi children uno per volta
    for(int j= 0; j<e1.children.length ; j++){
      
      // prendi il j-esimo child
      String child = e1.children[j]; 
     
     // in 'child' avro' il nome di uno dei children
      
      // fai un ciclo tra tutti i nodi per vedere se il child corrisponde a un certo nodo
      for(int k = 0; k<nodes.size(); k++){
        
          // prendi il k-esimo nodo
          Entity e2 = nodes.get( k );
          
          // se e' visibile e il suo nome e' uguale a quello del child
          if(e2.isVisible && e2.name.equals( child ) ){
            
            // --> sono padre-figlio e quindi devo creare una forza attrattiva tra di loro 
            
            // creo un vettore v
            PVector v = new PVector();
            
            // gli assegno la posizione di e2
            v.x = e2.position.x;
            v.y = e2.position.y;
            
            // gli sottraggo la posizione di e1
            v.sub( e1.position );
            
            // ne calcolo la lunghezza (e quindi la distanza)
            float magnitude = v.mag();
            
            
            // verifico se la loro distanza e' maggiore di 150
            // ( questo e' uno dei valori che potrebbero essere resi configurabili da interfaccia
            // permettendo ai padri di attrarre piu'/meno i figli
            if(magnitude > sogliaDistanzaPadreFiglio ){
              
              // lo divido per una certa costante, per causare un avvicinamento
              // progressivo (questo e' uno dei parametri da rendere configurabili da interfaccia
              // per causare maggiore/minore forza attrattiva)
              // la costante e' negativa per invertire la direzione del vettore, in modo da
              // tendere da e2 a e1
              v.mult( -1 * 0.001 );
            
              // aggiungo la forza ad e2
              e2.addForce( v );
              
              // lo inverto di nuovo (per -1) e lo diminuisco di nuovo
              // in modo che i padri siano attratti dai figli in misura
              // minore di quanto i figli sono attratti dai padri
              v.mult( -1 * 0.1 );
              
              // aggiungo la forza a e1
              e1.addForce( v );
              
            }
            
            
            // anche questo colore potrebbe essere reso configurabile
            stroke(113,126,60);
            noFill();
            
            // disegno la linea che collega padre e figlio visibili
            line(e1.position.x + globalOffsetX ,e1.position.y + globalOffsetY,  e2.position.x + globalOffsetX, e2.position.y +globalOffsetY);
          }
          
      }
      
    } 
    }//finisce if e1.isVisible
        
  }
  // crea le forze di attrazione tra i nodi ed i loro children - END
  
  //
  
  // disegna le relazioni di ogni genere e crea le forze di attrazione tra
  // i nodi coinvolti
  // prende le relazioni una per una
  // controlla che non siano tra quelle filtrate (quelle che non voglio vedere)
  // trova i nodi della relazione (e1 ed e2)
  // crea una forza di attrazione tra e1 ed e2
  // disegna la relazione (usando la sua configurazione)
  
  // prendo le relazioni una per una
  for(int i=0; i<relations.size(); i++){
    
    // prendo la i-esima relazione
    Relation r = relations.get( i );
    
    // in 'filteredRelations' sono contenuti i nomi dei tipi di
    // relazioni che non voglio vedere su schermo
    // --> controllo che il nome della relazione corrente (r) non sia
    // contenuto in 'filteredRelations'
    // --> uso 'indexOf'
    // data una stringa indexOf cerca al suo interno un'altra stringa
    // se la trova restituisce la posizione (casella/carattere) in cui l'ha trovata
    // se non la trova restituisce -1
    // stringa A = "mamma" e stringa B = "ma"
    // --> A.indexOf( B ) --> 0  (la prima casella [contando da 0] in cui trovo "ma")
    // se stringa B = "ps"
    // --> A.indexOf( B ) --> -1 (perche' "ps" non compare in A)
    // NOTA: nel primo caso (A="mamma" e B="ma") se voglio trovare la seconda
    // occorrenza di B posso fare cosi':
    // A.indexOf( B , 1) --> 3
    // quest'altra sintassi cerca a partire dalla casella indicata dal numero (1, in questo caso)
    
    // se "filteredRelations.indexOf( r.type )" restituisce -1 --> vuol dire che il
    // tipo di relazione corrente (r.type) non e' tra quelli filtrati
    // --> che posso elaborarlo
    if(filteredRelations.indexOf( r.type )==-1){
    
    // vuol dire che e' una relazione che devo elaborare
    
    // prendo tutti i nodi cercando r.node1 e r.node1
    
    // cerco il primo (loop tra tutti i nodi)
    for(int j=0; j<nodes.size(); j++){
      
      // prendo il nodo j-esimo
      Entity e1 = nodes.get(j);
      
      // e' visibile ed il suo nome e' uguale a r.node1
      if( e1.isVisible && e1.name.equals( r.node1 )  ){
          // we have found node 1
  
          // let's look for node 2!
          // loop tra i nodi per cercare il secondo
          for(int k = 0; k<nodes.size(); k++){
            
            // prendo il nodo k-esimo
            Entity e2 = nodes.get( k );
            
            // e' visibile e il suo nome e' uguale a r.node2
            if( e2.isVisible && e2.name.equals( r.node2 ) ){
              // we have found them both!
              // we can draw!
              
              
              // creo la forza di attrazione
              
              // creo un vettore
              PVector attraction =  new PVector();
              
              // ci copio dentro la posizione di e1
              attraction.x = e1.position.x;
              attraction.y = e1.position.y;
              
              // ci sottraggo la posizione di e2
              attraction.sub(e2.position);
              
              // calcolo la distanza
              float di = attraction.mag();
              
              
              // vedo se la distanza e' > di 150
              // (altro parametro da rendere configurabile tramite slider, per esempio)
              if(di>sogliaDistanzaRelazioni){
              
                // sono molto distanti, quindi desidero che si attraggano
                
                // inverto la direzione
                attraction.mult( -1 );
                
                // moltiplico per una costante per rallentare l'attrazione
                // (altro parametro da rendere configurabile: se lo aumento,
                // la forza di attrazione aumenta)
                attraction.mult( 0.01 );
                
                // applico la forza ad e1
                e1.addForce( attraction );
                
                // ne inverto la direzione
                attraction.mult( -1 );
                
                // applico la forza ad e2
                e2.addForce( attraction );
              
              }
              
              // usiamo il colore della configurazione della relazione
              // per impostare il colore del tratto
              stroke( r.configuration.r , r.configuration.g , r.configuration.b);
              noFill();
              
              // disegnamo la linea tra i due nodi
              line(  e1.position.x + globalOffsetX , e1.position.y + globalOffsetY, e2.position.x + globalOffsetX, e2.position.y  + globalOffsetY );
              
              if(r.direction.equals( "bidi" )){
                // devo disegnare freccette sul nodo 1
                pushMatrix();
                translate( e1.position.x + globalOffsetX, e1.position.y  + globalOffsetY );
                // rotate per allineare alla direzione della freccia
                //prendiamo l'asse X
                //PVector asseX = new PVector(0,1);
                //float angolo = PVector.angleBetween( attraction , asseX );
                float angolo = atan2(attraction.y,attraction.x);
                angolo = angolo + PI/2;
                fill( r.configuration.r , r.configuration.g , r.configuration.b);
                noStroke();
                rotate( angolo );
                // disegniamo un triangolino
                beginShape();
                vertex( 0,0 );
                vertex( 4,10 );
                vertex( -4,10 );
                endShape();
                popMatrix();
              } else if(r.direction.equals( "uni" ) || r.direction.equals( "bidi" )){
                // devo disegnare freccetta solo sul nodo 2 
                pushMatrix();
                translate( e2.position.x + globalOffsetX, e2.position.y  + globalOffsetY );
                // rotate per allineare alla direzione della freccia
                //prendiamo l'asse X
                //PVector asseX = new PVector(0,1);
                //float angolo = PVector.angleBetween( attraction , asseX );
                float angolo = atan2(attraction.y,attraction.x);
                angolo = angolo - PI/2;
                fill( r.configuration.r , r.configuration.g , r.configuration.b);
                noStroke();
                rotate( angolo );
                // disegniamo un triangolino
                beginShape();
                vertex( 0,0 );
                vertex( 4,10 );
                vertex( -4,10 );
                endShape();
                popMatrix();
              } 
              
              
              // se "r.configuration.icon" non e' null --> che e' stata
              // configurata un'icona
              if(r.configuration.icon!=null){
                
                // devo disegnare l'icona sulla linea della relazione
                
                // prendere posizione di e1
                PVector vettore = new PVector();
                vettore.x = e1.position.x;
                vettore.y = e1.position.y;
                
                // sottrarre la posizione di e2 --> differenza
                vettore.sub( e2.position );
                
                // in "offset" ho la distanza (in % rispetto alla lunghezza del
                // vettore differenza [quello che porta da e1 ad e2]), dal centro della
                // relazione, in cui devo disegrare l'icona
                // --> per capire in che punto devo disegnare l'icona della relazione
                // moltiplico il vettore differenza * (  0.5 +  offset)*)
                vettore.mult( (0.5 + r.configuration.offset)  );
                
                // gli inverto la direzione per andare da e1 a e2, invece che il contrario
                vettore.mult( -1 );
                
                //sommiamo al risultato la posizione di e1 --> e' dove disegno l'icona
                vettore.add( e1.position );
                
                // disegno l'icona
                // e la centro sul punto vettore.x,vettore.y
                // per centrarla uso la meta' della larghezza/altezza dell'icona
                
                // posso disegnare l'icona
                // invece di fare:
                // image( 
                //    r.configuration.icon , 
                //    vettore.x  - r.configuration.icon.width/2 , 
                //    vettore.y  - r.configuration.icon.height/2
                //);
                // aggiungo gli oggetti alla cache:
                DrawingCache elementoCache = new DrawingCache();
                elementoCache.immagine = r.configuration.icon;
                elementoCache.x = vettore.x  - r.configuration.icon.width/2;
                elementoCache.y = vettore.y  - r.configuration.icon.height/2;
                cache.add( elementoCache );
                
              }     
              
            }
          }
          
      }
    }
   }//r.type is not in filteredRelations
  }
  
  // disegna le relazioni di ogni genere e crea le forze di attrazione tra
  // i nodi coinvolti - END
  
  //disegnamo gli elementi della cache
  for(int i = 0; i<cache.size(); i++){
    DrawingCache elementoCache = cache.get( i );
    image( elementoCache.immagine , elementoCache.x  + globalOffsetX, elementoCache.y +globalOffsetY);
  }
  
  
  // disegna i nodi
  for(int i = 0; i<nodes.size(); i++){
    Entity e = nodes.get( i );
    // disegna i nodi solo se visibili
    if(e.isVisible){
      e.draw();
    }
  }
  // disegna i nodi - END
  
  
  
} // end of draw

// lo uso per memorizzare temporaneamente il nodo che sto trascinando
Entity isBeingMoved;


void handleMouse(){
  
  // se il mouse e' premuto --> vuol dire che potrei essere in mezzo
  // ad un drag'n drop
  if(mousePressed){
    
    
    // controllo se sto gia' trascinando un nodo
    if(isBeingMoved!=null){
      // I am already dragging a node
    } else {
      // I am not currently dragging a node
      // if it exists: I have to find the closest node and put it in isBeingMoved
      
      // prendo i nodi e vedo qual'e' il piu' vicino al mouse, entro una
      // certa distanza minima
      
      // prendo una Entity temporanea per fare le prove
      Entity testEntity = null;
      
      // prendo la posizione del mouse e la metto in un vettore
      PVector mousePosition = new PVector(mouseX - globalOffsetX ,mouseY - globalOffsetY );
      
      // imposto una variabile in cui memorizzero' la distanza mouse-nodo piu'
      // breve che ho trovato fino ad adesso
      // (nota: qualsiasi distanza su schermo sara' minore di 9999999
      // vuol dire che prendo anche la prima distanza, per fare i controlli,
      // anche se e' molto grande)
      float minimumDistance = 9999999;
     
       // prendo i nodi uno per uno 
      for(int i = 0; i<nodes.size(); i++){
        
        // prendo nodo i-esimo
        testEntity = nodes.get(i);
        
        // calcolo la distanza dal mouse
        PVector distance = new PVector();
        
        // ci copio la posizione del nodo
        distance.x = testEntity.position.x;
        distance.y = testEntity.position.y;
        
        // sottraggo la posizione del mouse --> ottengo vettore differenza
        distance.sub( mousePosition );
        
        // misuro vettore differenza con "mag()" --> ottengo la distanza
        float testDistance = distance.mag();
        
        // se la distanza e' <10 (la tolleranza di 10 pixel, in modo che
        // non devo essere troppo preciso a cliccare, e posso cliccare entro 10
        // pixel dal centro del nodo) e se e' minore di tutte le distanze che
        // ho trovato fino ad ora
        if(testDistance<10 && testDistance<minimumDistance){
          //println(testDistance);
          
          // la imposto come nuova distanza minima
          minimumDistance = testDistance;
          // so che fino ad adesso e' il nodo piu' papabile ad essere quello trascinato
         // --> lo metto dentro "isBeingMoved"  
          isBeingMoved = testEntity;
        } else {
          // in caso contrario non e' un nodo interessante ai fini del drag'n drop
          // --> lo metto a null
          testEntity = null;
        }
      }// finisce il ciclo
      
      
      // se "minimumDistance" e' rimasto uguale a "9999999"
      // --> vuol dire che non ho trovato nessun nodo entro 10 pixel dal
      // puntatore del mouse
      if(minimumDistance==9999999){
        // vuol dire che non ho cliccato su nessun nodo
        // e quindi imposto "isBeingMoved" a null 
        isBeingMoved = null;
        
        // aggiorno il globalOffsetX/Y per il pan
        
        globalOffsetX = globalOffsetX - (pmouseX-mouseX);
        globalOffsetY = globalOffsetY - (pmouseY-mouseY);
        
      }     
    }

    // a questo punto posso spostare il nodo di cui sto facendo drag'n drop

    // se "isBeingMoved" non e' uguale a null --> vuol dire che sto facendo
    // un drag'n drop
    if(isBeingMoved!=null){
      // ok, I'm clicking on a node
      // --> gli posso assegnare la posizione del mouse
      isBeingMoved.position.x = mouseX - globalOffsetX;
      isBeingMoved.position.y = mouseY - globalOffsetY;
    }    
  } else {
    // the user is not dragging any more
    // so I put isBeingMoved to null
    isBeingMoved = null;
  }
}





public void handleSliderEvents(GValueControl slider, GEvent event) {
    //println("integer value:" + slider.getValueI() + " float value:" + slider.getValueF());
    
    if( slider.tag.equals( "sogliaDistanzaPadreFiglio" ) ){
      sogliaDistanzaPadreFiglio = slider.getValueF();
    }
    else if( slider.tag.equals( "sogliaDistanzaRelazioni" ) ){
      sogliaDistanzaRelazioni = slider.getValueF();
    }

}
